@@ -70,16 +70,14 @@ xsltNamespaceAlias(xsltStylesheetPtr style, xmlNodePtr node) {
     sprefix = xsltGetNsProp(node, (const xmlChar *)"stylesheet-prefix",
 	                   XSLT_NAMESPACE);
     if (sprefix == NULL) {
-	xsltPrintErrorContext(NULL, style, node);
-	xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, style, node,
 	    "namespace-alias: stylesheet-prefix attribute missing\n");
 	return;
     }
     rprefix = xsltGetNsProp(node, (const xmlChar *)"result-prefix",
 	                   XSLT_NAMESPACE);
     if (rprefix == NULL) {
-	xsltPrintErrorContext(NULL, style, node);
-	xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, style, node,
 	    "namespace-alias: result-prefix attribute missing\n");
 	goto error;
     }
@@ -89,8 +87,7 @@ xsltNamespaceAlias(xsltStylesheetPtr style, xmlNodePtr node) {
 	sNs = xmlSearchNs(node->doc, node, sprefix);
     }
     if ((sNs == NULL) || (sNs->href == NULL)) {
-	xsltPrintErrorContext(NULL, style, node);
-	xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, style, node,
 	    "namespace-alias: prefix %s not bound to any namespace\n",
 	                 sprefix);
 	goto error;
@@ -101,8 +98,7 @@ xsltNamespaceAlias(xsltStylesheetPtr style, xmlNodePtr node) {
 	rNs = xmlSearchNs(node->doc, node, rprefix);
     }
     if ((rNs == NULL) || (rNs->href == NULL)) {
-	xsltPrintErrorContext(NULL, style, node);
-	xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, style, node,
 	    "namespace-alias: prefix %s not bound to any namespace\n",
 	                 rprefix);
 	goto error;
@@ -110,8 +106,7 @@ xsltNamespaceAlias(xsltStylesheetPtr style, xmlNodePtr node) {
     if (style->nsAliases == NULL)
 	style->nsAliases = xmlHashCreate(10);
     if (style->nsAliases == NULL) {
-	xsltPrintErrorContext(NULL, style, node);
-	xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, style, node,
 	    "namespace-alias: cannot create hash table\n");
 	goto error;
     }
@@ -125,6 +120,161 @@ error:
 	xmlFree(rprefix);
 }
 
+/**
+ * xsltNsInScope:
+ * @doc:  the document
+ * @node:  the current node
+ * @ancestor:  the ancestor carrying the namespace
+ * @prefix:  the namespace prefix
+ *
+ * Copy of xmlNsInScope which is not public ...
+ * 
+ * Returns 1 if true, 0 if false and -1 in case of error.
+ */
+static int
+xsltNsInScope(xmlDocPtr doc ATTRIBUTE_UNUSED, xmlNodePtr node,
+             xmlNodePtr ancestor, const xmlChar * prefix)
+{
+    xmlNsPtr tst;
+
+    while ((node != NULL) && (node != ancestor)) {
+        if ((node->type == XML_ENTITY_REF_NODE) ||
+            (node->type == XML_ENTITY_NODE) ||
+            (node->type == XML_ENTITY_DECL))
+            return (-1);
+        if (node->type == XML_ELEMENT_NODE) {
+            tst = node->nsDef;
+            while (tst != NULL) {
+                if ((tst->prefix == NULL)
+                    && (prefix == NULL))
+                    return (0);
+                if ((tst->prefix != NULL)
+                    && (prefix != NULL)
+                    && (xmlStrEqual(tst->prefix, prefix)))
+                    return (0);
+                tst = tst->next;
+            }
+        }
+        node = node->parent;
+    }
+    if (node != ancestor)
+        return (-1);
+    return (1);
+}
+
+/**
+ * xsltSearchPlainNsByHref:
+ * @doc:  the document
+ * @node:  the current node
+ * @href:  the namespace value
+ *
+ * Search a Ns aliasing a given URI and without a NULL prefix.
+ * Recurse on the parents until it finds
+ * the defined namespace or return NULL otherwise.
+ * Returns the namespace pointer or NULL.
+ */
+static xmlNsPtr
+xsltSearchPlainNsByHref(xmlDocPtr doc, xmlNodePtr node, const xmlChar * href)
+{
+    xmlNsPtr cur;
+    xmlNodePtr orig = node;
+
+    if ((node == NULL) || (href == NULL))
+        return (NULL);
+
+    while (node != NULL) {
+        if ((node->type == XML_ENTITY_REF_NODE) ||
+            (node->type == XML_ENTITY_NODE) ||
+            (node->type == XML_ENTITY_DECL))
+            return (NULL);
+        if (node->type == XML_ELEMENT_NODE) {
+            cur = node->nsDef;
+            while (cur != NULL) {
+                if ((cur->href != NULL) && (cur->prefix != NULL) &&
+		    (href != NULL) && (xmlStrEqual(cur->href, href))) {
+		    if (xsltNsInScope(doc, orig, node, cur->href) == 1)
+			return (cur);
+                }
+                cur = cur->next;
+            }
+            if (orig != node) {
+                cur = node->ns;
+                if (cur != NULL) {
+                    if ((cur->href != NULL) && (cur->prefix != NULL) &&
+		        (href != NULL) && (xmlStrEqual(cur->href, href))) {
+			if (xsltNsInScope(doc, orig, node, cur->href) == 1)
+			    return (cur);
+                    }
+                }
+            }    
+        }
+        node = node->parent;
+    }
+    return (NULL);
+}
+
+/**
+ * xsltGetPlainNamespace:
+ * @ctxt:  a transformation context
+ * @cur:  the input node
+ * @ns:  the namespace
+ * @out:  the output node (or its parent)
+ *
+ * Find the right namespace value for this prefix, if needed create
+ * and add a new namespace decalaration on the node
+ * Handle namespace aliases and make sure the prefix is not NULL, this
+ * is needed for attributes.
+ *
+ * Returns the namespace node to use or NULL
+ */
+xmlNsPtr
+xsltGetPlainNamespace(xsltTransformContextPtr ctxt, xmlNodePtr cur,
+                      xmlNsPtr ns, xmlNodePtr out) {
+    xsltStylesheetPtr style;
+    xmlNsPtr ret;
+    const xmlChar *URI = NULL; /* the replacement URI */
+
+    if ((ctxt == NULL) || (cur == NULL) || (out == NULL) || (ns == NULL))
+	return(NULL);
+
+    style = ctxt->style;
+    while (style != NULL) {
+	if (style->nsAliases != NULL)
+	    URI = (const xmlChar *) xmlHashLookup(style->nsAliases, ns->href);
+	if (URI != NULL)
+	    break;
+
+	style = xsltNextImport(style);
+    }
+
+    if (URI == NULL)
+	URI = ns->href;
+
+    if ((out->parent != NULL) &&
+	(out->parent->type == XML_ELEMENT_NODE) &&
+	(out->parent->ns != NULL) &&
+	(out->parent->ns->prefix != NULL) &&
+	(xmlStrEqual(out->parent->ns->href, URI)))
+	ret = out->parent->ns;
+    else {
+	if (ns->prefix != NULL) {
+	    ret = xmlSearchNs(out->doc, out, ns->prefix);
+	    if ((ret == NULL) || (!xmlStrEqual(ret->href, URI)) ||
+	        (ret->prefix == NULL)) {
+		ret = xsltSearchPlainNsByHref(out->doc, out, URI);
+	    }
+	} else {
+	    ret = xsltSearchPlainNsByHref(out->doc, out, URI);
+	}
+    }
+
+    if (ret == NULL) {
+	if (out->type == XML_ELEMENT_NODE)
+	    ret = xmlNewNs(out, URI, ns->prefix);
+    }
+    return(ret);
+}
+
 /**
  * xsltGetSpecialNamespace:
  * @ctxt:  a transformation context
@@ -148,21 +298,35 @@ xsltGetSpecialNamespace(xsltTransformContextPtr ctxt, xmlNodePtr cur,
     if ((ctxt == NULL) || (cur == NULL) || (out == NULL) || (URI == NULL))
 	return(NULL);
 
+    if ((prefix == NULL) && (URI[0] == 0)) {
+	ret = xmlSearchNs(out->doc, out, NULL);
+	if (ret != NULL) {
+	    ret = xmlNewNs(out, URI, prefix);
+	    return(ret);
+	}
+	return(NULL);
+    }
+
     if ((out->parent != NULL) &&
 	(out->parent->type == XML_ELEMENT_NODE) &&
 	(out->parent->ns != NULL) &&
 	(xmlStrEqual(out->parent->ns->href, URI)))
 	ret = out->parent->ns;
-    else
+    else 
 	ret = xmlSearchNsByHref(out->doc, out, URI);
 
-    if (ret == NULL) {
+    if ((ret == NULL) || (ret->prefix == NULL)) {
 	if (prefix == NULL) {
 	    do {
 		sprintf(nprefix, "ns%d", prefixno++);
 		ret = xmlSearchNs(out->doc, out, (xmlChar *)nprefix);
 	    } while (ret != NULL);
 	    prefix = (const xmlChar *) &nprefix[0];
+	} else if ((ret != NULL) && (ret->prefix == NULL)) {
+	    /* found ns but no prefix - search for the prefix */
+	    ret = xmlSearchNs(out->doc, out, prefix);
+	    if (ret != NULL)
+	        return(ret);	/* found it */
 	}
 	if (out->type == XML_ELEMENT_NODE)
 	    ret = xmlNewNs(out, URI, prefix);
@@ -197,7 +361,7 @@ xsltGetNamespace(xsltTransformContextPtr ctxt, xmlNodePtr cur, xmlNsPtr ns,
     while (style != NULL) {
 	if (style->nsAliases != NULL)
 	    URI = (const xmlChar *) 
-		xmlHashLookup(ctxt->style->nsAliases, ns->href);
+		xmlHashLookup(style->nsAliases, ns->href);
 	if (URI != NULL)
 	    break;
 
@@ -212,8 +376,16 @@ xsltGetNamespace(xsltTransformContextPtr ctxt, xmlNodePtr cur, xmlNsPtr ns,
 	(out->parent->ns != NULL) &&
 	(xmlStrEqual(out->parent->ns->href, URI)))
 	ret = out->parent->ns;
-    else
-	ret = xmlSearchNsByHref(out->doc, out, URI);
+    else {
+	if (ns->prefix != NULL) {
+	    ret = xmlSearchNs(out->doc, out, ns->prefix);
+	    if ((ret == NULL) || (!xmlStrEqual(ret->href, URI))) {
+		ret = xmlSearchNsByHref(out->doc, out, URI);
+	    }
+	} else {
+	    ret = xmlSearchNsByHref(out->doc, out, URI);
+	}
+    }
 
     if (ret == NULL) {
 	if (out->type == XML_ELEMENT_NODE)
@@ -237,7 +409,7 @@ xsltGetNamespace(xsltTransformContextPtr ctxt, xmlNodePtr cur, xmlNsPtr ns,
 xmlNsPtr
 xsltCopyNamespaceList(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	              xmlNsPtr cur) {
-    xmlNsPtr ret = NULL;
+    xmlNsPtr ret = NULL, tmp;
     xmlNsPtr p = NULL,q;
     const xmlChar *URI;
 
@@ -255,12 +427,22 @@ xsltCopyNamespaceList(xsltTransformContextPtr ctxt, xmlNodePtr node,
     while (cur != NULL) {
 	if (cur->type != XML_NAMESPACE_DECL)
 	    break;
+
+	/*
+	 * Avoid duplicating namespace declrations on the tree
+	 */
 	if ((node != NULL) && (node->ns != NULL) &&
             (xmlStrEqual(node->ns->href, cur->href)) &&
             (xmlStrEqual(node->ns->prefix, cur->prefix))) {
 	    cur = cur->next;
 	    continue;
 	}
+	tmp = xmlSearchNs(node->doc, node, cur->prefix);
+	if ((tmp != NULL) && (xmlStrEqual(tmp->href, cur->href))) {
+	    cur = cur->next;
+	    continue;
+	}
+	
 	if (!xmlStrEqual(cur->href, XSLT_NAMESPACE)) {
 	    /* TODO apply cascading */
 	    URI = (const xmlChar *) xmlHashLookup(ctxt->style->nsAliases,
